#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif //HAVE_CONFIG_H

#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>

#include <taglib/fileref.h>
#include <glibmm.h>

#include <boost/format.hpp>
static boost::format fsizefmt ("%llu");

// Plugin-specific include
#include <taglib/taglib.h>
#include <taglib/fileref.h>
#include <taglib/tfile.h>
#include <taglib/tag.h>

#include <taglib/mpcfile.h>
#include <taglib/apetag.h>
#include <taglib/apeitem.h>
#include <taglib/apefooter.h>

#include "bmp/database-types.hh"
#include "bmp/library-types.hh"
#include "bmp/library-ops.hh"

using namespace Bmp;
using namespace TagLib;
using namespace Glib;

extern "C" int  _plugin_has_accessors;
int  _plugin_has_accessors = 1;

extern "C" int  _plugin_version;
int _plugin_version = PLUGIN_VERSION;

extern "C" const char ** _mimetypes ()
{
  static const char * _types[] =
  {
    "application/x-apetag",
    NULL
  };
  return _types;
}

extern "C" bool _set (std::string const& name, Track & track)
{
  MPC::File opfile (name.c_str());
  if (!(opfile.isOpen() && opfile.isValid()))
    return false;

  APE::Tag * tag = opfile.APETag(true);
  if (tag)
  {
      if (track.mb_album_artist)
        tag->addValue ("Album Artist",
                        String (track.mb_album_artist.get(),            String::UTF8));

      if (track.mb_album_artist_id)
        tag->addValue ("MUSICBRAINZ_ALBUMARTISTID",
                        String (track.mb_album_artist_id.get(),         String::UTF8));

      if (track.mb_album_artist_sort_name)
        tag->addValue ("ALBUMARTISTSORT",
                        String (track.mb_album_artist_sort_name.get(),  String::UTF8));

      if (track.mb_track_id)
        tag->addValue ("MUSICBRAINZ_TRACKID",
                        String (track.mb_track_id.get(),                String::UTF8));

      if (track.mb_album_id)
        tag->addValue ("MUSICBRAINZ_ALBUMID",
                        String(track.mb_album_id.get(),                 String::UTF8));

      if (track.mb_artist_id)
        tag->addValue ("MUSICBRAINZ_ARTISTID",
                        String (track.mb_artist_id.get(),               String::UTF8));


      if (track.mb_artist_sort_name)
        tag->addValue ("ARTISTSORT",
                        String (track.mb_artist_sort_name.get(),        String::UTF8));

      if (track.mb_release_date)
        tag->addValue ("Year",
                        String (track.mb_release_date.get(),            String::UTF8));

      if (track.asin)
        tag->addValue ("ASIN",
                        String (track.asin.get(),                       String::UTF8));

      if (track.asin)
        tag->addValue ("MUSICIP_PUID",
                        String (track.asin.get(),                       String::UTF8));

      opfile.save();

#if 0
      tag = opfile.APETag ();
      struct stat tagstat;
      stat (name.c_str(), &tagstat);
      std::string size_str = (fsizefmt % (unsigned long long int) (tagstat.st_size)).str ();
      TagLib::ByteVector tagData (tag->render());
      track.hash = Util::md5_hex_string (tagData.data (), tagData.size ()) + size_str;
#endif

      return true;
  }
  else
  {
      return false;
  }
}

extern "C" bool _get (TagLib::File * p, DB::Row & row, std::string const& name)
{
  struct {
      AttributeId     datum;
      char          * id;
  } mb_metadata_ape[] = {
    { ATTRIBUTE_MB_ALBUM_ARTIST,                "Album Artist"                },
    { ATTRIBUTE_MB_ALBUM_ARTIST_ID,             "MUSICBRAINZ_ALBUMARTISTID"   },
    { ATTRIBUTE_MB_ALBUM_ARTIST_SORTNAME,       "ALBUMARTISTSORT"             }, 
    { ATTRIBUTE_MB_TRACK_ID,                    "MUSICBRAINZ_TRACKID"         },
    { ATTRIBUTE_MB_ALBUM_ID,                    "MUSICBRAINZ_ALBUMID"         },
    { ATTRIBUTE_MB_ARTIST_ID,                   "MUSICBRAINZ_ARTISTID"        },
    { ATTRIBUTE_MB_ARTIST_SORTNAME,             "ARTISTSORT"                  },
    { ATTRIBUTE_MB_RELEASE_DATE,                "Year"                        },
    { ATTRIBUTE_ASIN,                           "ASIN"                        },
    { ATTRIBUTE_MUSICIP_PUID,                   "MUSICIP_PUID"                },
  };


  MPC::File *file;

  if (!(file = dynamic_cast<MPC::File*> (p)))
    return false;

  APE::Tag * tag = file->APETag();

  if (tag)
  {
      TagLib::APE::ItemListMap const& items (tag->itemListMap());

      for (unsigned int n = 0; n < G_N_ELEMENTS (mb_metadata_ape); ++n)
      {
        if (!items[mb_metadata_ape[n].id].isEmpty())
        {
          row.insert (std::make_pair (get_attribute_info(mb_metadata_ape[n].datum).id,
              ustring (items[mb_metadata_ape[n].id].toString().toCString (true))));
        }
      }

#if 0
      struct stat tagstat;
      stat(name.c_str(), &tagstat);
      std::string size_str = (fsizefmt % (unsigned long long int) (tagstat.st_size)).str ();
      TagLib::ByteVector tagData (tag->render());
      std::string hash = Util::md5_hex_string (tagData.data (), tagData.size ()) + size_str;
      row.insert (std::make_pair (get_attribute_info(ATTRIBUTE_HASH).id, Glib::ustring(hash)));
#endif

      return true;
    }
  else
    {
      return false;
    }
}
