//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2007 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non-GPL compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#ifndef BMP_STREAMS_SHOUTCAST_HH
#define BMP_STREAMS_SHOUTCAST_HH

#include <glibmm/ustring.h>
#include <gtkmm/liststore.h>
#include <gtkmm/treemodelfilter.h>
#include <gtkmm/treeview.h>
#include <libglademm/xml.h>
#include <sigc++/signal.h>
#include "minisoup.hh"
#include "xml.hh"

#include "radio-directory-types.hh"

namespace Bmp
{
  namespace RadioDirectory
  {
    class Shoutcast
      : public Gtk::TreeView
    {
      public:

        typedef sigc::signal <void, StreamListT const&> SignalListUpdated;

        Shoutcast (BaseObjectType                       * obj,
                   Glib::RefPtr<Gnome::Glade::Xml> const& xml);

        virtual ~Shoutcast ()
        {
          g_hash_table_destroy (m_cache);
        }

        void refresh (bool force = false);

        SignalListUpdated&
        s_list_updated ()
        {
          return s_list_updated_;
        }

        SignalStartStop&
        s_start ()
        { return s_start_; }

        SignalStartStop&
        s_stop ()
        {
          return s_stop_;
        }

      protected:

        void on_row_activated (Gtk::TreeModel::Path const& path, Gtk::TreeViewColumn * column);

      private:

        SignalListUpdated s_list_updated_;
        SignalStartStop   s_start_;
        SignalStartStop   s_stop_;

        class Columns
          : public Gtk::TreeModel::ColumnRecord
        {
          public:
            Gtk::TreeModelColumn<Glib::ustring> name;
            Columns ()
            {
              add (name);
            }
        };

        Columns genres;

        void  genre_cell_data_func (Gtk::CellRenderer *cell, Gtk::TreeIter const& iter);
        void  build_genre_list ();
        void  refresh_wrapper ();
        void  refresh_parse_and_emit_updated (xmlDocPtr doc);
        void  refresh_callback (char const* data, guint size, guint code, std::string const& genre);

        Glib::RefPtr<Gtk::ListStore> m_genres;
        Gtk::TreeModel::iterator m_current_selected_iter;
        GHashTable* m_cache;
        StreamListT m_stream_list;
        Soup::RequestRefP  m_refresh_request; 
    }; // end class Shoutcast

  } // end namespace RadioDirectory
} // end namespace Bmp

#endif // !BMP_STREAMS_SHOUTCAST_HH
